/*-------------------------------------------------------------------*/
/* TRAIN.C         Pasi Frnti                                       */
/*                 Timo Kaukoranta                                   */
/*                 Juha Kivijrvi                                    */
/*                                                                   */
/* Generate the training set for VQ algorithm.                       */
/* Supports:                                                         */
/*     - PBM-images: Original                                        */
/*     - PGM-images: Original, BTC-binarized, Minmax-normalized      */
/*     - PPM-images: 1x1 RGB-tuples                                  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "TRAIN"
#define  VersionNumber  "Version 0.10"
#define  LastUpdated    "24.2.99"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define  FACTFILE "train.fac"
#include "cb.h"
#include "pgm.h"
#include "file.h"
#include "image.h"
#include "parametr.c"
#include "owntypes.h"
#include "memctrl.h"
#include "bintree.h"
#include "random.h"

/*-----------------------  Global variables ---------------------------*/

int             NumberOfImages = 0;
int             Xsize, Ysize;
int             Vsize;
FILETYPE        InputFileType;

#define         InitialSize           1024
#define         MinMaxScalingConstant 1000000
#define         MeanSdConstantMean    500000
#define         MeanSdConstantSd      100000


/* -------------------------- Misc -----------------------------------*/

#if defined(max) || defined(min)
#undef max
#undef min
#endif

#define  max(a,b) ((a) > (b) ? (a) : (b))
#define  min(a,b) ((a) < (b) ? (a) : (b))


/*=======================  Time routines  ============================*/


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t  tmp;
  int     hours, minutes, seconds;
  long    elapsed;

  elapsed  = (long)time(&tmp) - start;
  hours    = (int)(elapsed / 3600);
  elapsed  = (int)(elapsed % 3600);
  minutes  = (int)(elapsed / 60);
  seconds  = (int)(elapsed % 60);

  switch( Value(QuietLevel) )
    {
    case 0:  break;
    case 1:  printf(" %i:%02i:%02i", hours, minutes, seconds);
             break;
    default: printf("Time: %i:%02i:%02i\n", hours, minutes, seconds);
             break;
    }
}


/*======================  Printing routines  ========================*/


void PrintInfo(void)
{
  printf("%s     %s                (last modified %s)\n",ProgName,VersionNumber,LastUpdated);
  printf("Usage: %s outputfile inputfiles [%coptions]\n", ProgName, OPTION_SYMBOL);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s     %s\n",ProgName,VersionNumber);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char*  OutFileName,
                               char   TSFileNames[21][80])
{
  int i;

  if( Value(QuietLevel) >= 1 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("\nGenerating training set %s using %i %s:\n", OutFileName, NumberOfImages, NumberOfImages>1 ? "images" : "image");
    printf("Training set file: %s\n", OutFileName);
    printf("Input images:\n");
    for(i=1; i<=NumberOfImages; i++)
       {
       printf("%s\n", TSFileNames[i]);
       }
    printf("\n");
    PrintSelectedOptions();
    printf("Vector dimensions: (%i, %i) \n",Xsize,Ysize);
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


void PrintProgress(int y, int total)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("%5.1f %% done.\r", (double)(y) * 100 / total);
    fflush(stdout);
    }
}


/*-------------------------------------------------------------------*/


void PrintTSstatistics(TRAININGSET* TS)
{
  printf("Training set size     %d\n", BookSize(TS));
  printf("Total frequency       %d\n", TotalFreq(TS));
  printf("Maximal value         %d\n", TS->MaxValue);
  printf("Vector size           %d\n", VectorSize(TS));
}  
  

/*=======================  E N T R O P Y  ============================*/


static double Entropy(float p)
{
  double e = ( p > 0 ? -p * log(p) / log(2) : 0 );

  return( e );
}


/*-------------------------------------------------------------------*/


void CalculateEntropy(TRAININGSET* TS)
{
  double e, p;
  double entropy=0;
  int    i;

  for(i=0; i<TS->CodebookSize; i++)
     {
     p = (double) VectorFreq(TS, i) / TS->TotalFreq;
     e = Entropy(p);
     entropy += e;
     }

  printf("\nEntropy of the training set = %6.3f (bits/vector) \n", entropy);
}


/*=================== Construction of the vector ========================*/


static void ConstructSpatialVector(IMAGE* Image, int StartX, int StartY,
                                   VECTORTYPE v)
{
  int    x, y, i=0;

  for( y=StartY; y<=StartY+Ysize-1; y++ )
     for( x=StartX; x<=StartX+Xsize-1; x++ )
         {
         v[i++] = GetImageGray(Image,x,y);
         }
}


/*-------------------------------------------------------------------*/


static void ConstructColorVector(IMAGE* Image, int StartX, int StartY,
                                 VECTORTYPE v)
{
  v[0] = GetImageRed(Image,StartX,StartY);
  v[1] = GetImageGreen(Image,StartX,StartY);
  v[2] = GetImageBlue(Image,StartX,StartY);

  if( Vsize != 3 )
    {
    printf("\nERROR: Wrong vectorsize for RGB images. Block size = %ix%i, vector size = %i.\n",
           Xsize, Ysize, Vsize);
    exit(-1);
    }
}


/*-------------------------------------------------------------------*/


static void ConstructVector(IMAGE* Image, int x, int y, VECTORTYPE v)
{
  if( InputFileType == PPM )
    {
    ConstructColorVector(Image, x, y, v);
    }
  else
    {
    ConstructSpatialVector(Image, x, y, v);
    }
}


/*=======================  P R O C E S S  ==========================*/


int CompareNodes(void* a, void* b, void* info)
{
  return ( CompareVectors( ((BOOKNODE*) a)->vector,
                           ((BOOKNODE*) b)->vector, *((int*)info)) );
}    


/* ----------------------------------------------------------------- */


static void InsertVectorToBintree(BINTREE* Tree, VECTORTYPE v)
{
  BOOKNODE* node;
  void*     found;

  node  = allocate(sizeof(BOOKNODE));
  *node = CreateEmptyNode(Vsize);
  CopyVector(v, node->vector, Vsize);
  node->freq = 1;

  found = InsertToBintreeNoDuplicates( Tree, (void*) node, (void*)(&Vsize)); 
  if (found)
    {
    ((BOOKNODE*) found)->freq++;
    FreeNode(*node);
    deallocate(node);
    }
}    


/*-------------------------------------------------------------------*/


static void InsertNodeToTrainingSet(TRAININGSET* TS, BOOKNODE* node)
{
  CopyNode(node, &Node(TS, BookSize(TS)), Vsize);
  TotalFreq(TS) += node->freq;
  BookSize(TS)++;
}    


/*-------------------------------------------------------------------*/


static void CreateTrainingSet(TRAININGSET* TS, BINTREE* Tree)
{
  STACK     S;
  BOOKNODE* node;
  
  InitInOrderBintree(Tree, &S);
  
  node = (BOOKNODE*) InOrderBintree(&S);
  while (node != NULL)
    {
    InsertNodeToTrainingSet(TS, node);
    FreeNode(*node);
    deallocate(node);
    node = (BOOKNODE*) InOrderBintree(&S);
    }
}


/*-------------------------------------------------------------------*/


static void CalculateStatistics(VECTORTYPE v, 
                                double*    mean, 
                                double*    sd,
                                int*       minimum, 
                                int*       maximum)
{
  llong  squaresum = 0;
  llong  sum = 0;
  int    Min = v[0];
  int    Max = v[0];
  int    i;
  double pmean, sqmean, var;

  for ( i=0; i<Vsize; i++ )
    {
    sum += (llong) v[i];
    squaresum += (llong) v[i]*v[i];
    Min = min(Min, v[i]);
    Max = max(Max, v[i]);
    }
  pmean      = (double) sum / Vsize;
  sqmean     = (double) squaresum / Vsize;
  var        = sqmean - pmean*pmean;
  (*sd)      = (var>0 ? sqrt(var) : 0);
  (*mean)    = pmean;
  (*minimum) = Min;
  (*maximum) = Max;
}


/*-------------------------------------------------------------------*/


static void LocalBinarization(VECTORTYPE v, double mean) 
{
  int i;

  for ( i=0; i<Vsize; i++ )
    {
    v[i] = (v[i] >= mean) ? 1 : 0;
    }
}


/*-------------------------------------------------------------------*/


void MeanSdNormalization(VECTORTYPE v, double mean, double sd)
{
  int i;

  for ( i=0; i<Vsize; i++ )
    {
    v[i] = sd > 0 ? roundint( ((double) MeanSdConstantMean +               
           MeanSdConstantSd * (v[i]-mean) / sd) ) : MeanSdConstantMean; 
    if (v[i] < 0) v[i] = 0;
    if (v[i] > 2*MeanSdConstantMean) v[i] = 2*MeanSdConstantMean;
    }
}


/*-------------------------------------------------------------------*/


void MinMaxNormalization(VECTORTYPE v, int Min, int Max)
{
  int i;
  int range = Max-Min;

  for ( i=0; i<Vsize; i++ )
    {
    v[i] = range ? roundint( ((double) MinMaxScalingConstant*(v[i]-Min) /  
           range) ) : MinMaxScalingConstant/2;                            
    }
}

/* ................................................................. */
/* Example (min=3, max=8)
        range      5
        min..max  3    4    5     6      7    8
        x - min   0    1    2     3      4    5
        / range   0.0  0.2    0.4  0.6  0.8  1.0
        * 255      0    51    102  153  204  255           */
/* ................................................................. */


/*-------------------------------------------------------------------*/


static void QuantizeBlock(VECTORTYPE v)
{
  int    mask[9] = { 0x00, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff };
  int    q = Value(PreQuantization);
  int    i;

  for ( i=0; i<Vsize; i++ )
    v[i] &= mask[q];
}


/*-------------------------------------------------------------------*/


static void ProcessBlock(VECTORTYPE v)
{
  double  mean, sd;
  int     Min, Max;

  CalculateStatistics(v, &mean, &sd, &Min, &Max);

  if( Value(PreQuantization) != Default(PreQuantization) )
     {
     QuantizeBlock(v);
     }

  switch(Value(Preprocessing))
     {
     case LOCBIN:  LocalBinarization(v,mean);      break;
     case MEAN_SD: MeanSdNormalization(v,mean,sd); break;
     case MIN_MAX: MinMaxNormalization(v,Min,Max); break; 
     default:      break;
     }
}


/*-------------------------------------------------------------------*/


static void ProcessImage(IMAGE* Image, BINTREE* Tree)
{
  VECTORTYPE  v;
  int x, y;
  int Ystep = (InputFileType == PPM) ? 1 : Ysize;

  v = CreateEmptyVector(Vsize);

  for( y=1; y<=Image->ImageSizeY; y+=Ystep )
     {
     ReadNewBuffer(Image);
     for( x=1; x<=Image->ImageSizeX; x+=Xsize )
        {
        ConstructVector(Image, x, 1, v); 
        ProcessBlock(v);
        InsertVectorToBintree(Tree, v);
        }
     PrintProgress(y, Image->ImageSizeY);
     }

  if( (Value(QuietLevel)) >= 2 && (NumberOfImages>1) )
    {
    printf("Training set contains now %u cells.\n", BintreeSize(Tree));
    }

  FreeVector(v);
}


/*======================= Parameter parsing =========================*/


void ParameterParsing(int   argc, 
                      char* argv[], 
                      char* OutFileName,
                      char  TrainingImage[21][80])
{                    
  ParameterInfo paraminfo[21] =  { { OutFileName, FormatNameTS, 0, OUTFILE }, 
                                   { TrainingImage[1],  "",     0, INFILE  },
                                   { TrainingImage[2],  "",     1, INFILE  },
                                   { TrainingImage[3],  "",     2, INFILE  },
                                   { TrainingImage[4],  "",     3, INFILE  },
                                   { TrainingImage[5],  "",     4, INFILE  },
                                   { TrainingImage[6],  "",     5, INFILE  },
                                   { TrainingImage[7],  "",     6, INFILE  },
                                   { TrainingImage[8],  "",     7, INFILE  },
                                   { TrainingImage[9],  "",     8, INFILE  },
                                   { TrainingImage[10], "",     9, INFILE  },
                                   { TrainingImage[11], "",    10, INFILE  },
                                   { TrainingImage[12], "",    11, INFILE  },
                                   { TrainingImage[13], "",    12, INFILE  },
                                   { TrainingImage[14], "",    13, INFILE  },
                                   { TrainingImage[15], "",    14, INFILE  },
                                   { TrainingImage[16], "",    15, INFILE  },
                                   { TrainingImage[17], "",    16, INFILE  },
                                   { TrainingImage[18], "",    17, INFILE  },
                                   { TrainingImage[19], "",    18, INFILE  },  
                                   { TrainingImage[20], "",    19, INFILE  } };

  NumberOfImages = ParseParameters( argc, argv, 21, paraminfo) - 1;
}


/*-------------------------------------------------------------------*/


void InitializeTrainingSet(TRAININGSET* TS, int size, int maxvalue)
{
  int  maxval;
  int  bpe;
  char GenMethod[MaxGenMethodLength];
  
  strcpy(GenMethod, ProgName);
  strcat(GenMethod, " ");
  strcat(GenMethod, VersionNumber);
  switch (Value(Preprocessing))
    {
    case LOCBIN:  { maxval = 1;                     bpe = 1; break; }
    case MIN_MAX: { maxval = MinMaxScalingConstant; bpe = 4; break; }
    case MEAN_SD: { maxval = 2*MeanSdConstantMean;  bpe = 4; break; }
    default:      { maxval = maxvalue;              bpe = 1; break; }
    }

  CreateNewTrainingSet(TS, size, Xsize, Ysize, bpe,
                       0, maxval, Value(Preprocessing), GenMethod);
  BookSize(TS) = 0;
}


/*========================== Main program ===========================*/


void main(int argc, char* argv[])
{
  char           TSFileName[21][80];
  char           OutputFileName[80];
  FILE*          InputFile;
  IMAGE          InputImage;
  TRAININGSET    TS;
  long           Watch;
  int            i;
  BINTREE        Tree;
  int            maxval = 0;

  ParameterParsing( argc, argv, OutputFileName, TSFileName);

  InputFileType = DetermineFileFormat(TSFileName[1]);

  if (InputFileType==PPM) 
    { 
    Xsize = 1; 
    Ysize = 3; 
    }
  else 
    { 
    Xsize = Value(Xsize); 
    Ysize = Value(Ysize); 
    }
  Vsize = Xsize * Ysize;
 
  PrintOperatingInfo(OutputFileName, TSFileName);
  SetWatch(&Watch);

  InitBintree(&Tree, &CompareNodes);

  for(i=1; i<=NumberOfImages; i++)
     {
     InputFile = OpenFile(TSFileName[i], INPUT,  NO);
     InitializeImage(InputFile, &InputImage, INPUT, BLOCKBUFFER, 
                     (InputFileType==PPM? 1 : Ysize));
     if (i == 1)  maxval = InputImage.MaxValue;
     if( Value(QuietLevel) >= 2 )
        {
        printf("Generating training set from %s.\n", TSFileName[i]);
        }
     ProcessImage(&InputImage, &Tree);
     FreeImage(&InputImage);
     fclose(InputFile);
     }

  InitializeTrainingSet(&TS, BintreeSize(&Tree), maxval); 
  CreateTrainingSet(&TS, &Tree);
  if (Value(RandomizeOrder))
    ShuffleTS(&TS);
  PrintTSstatistics(&TS);
  CalculateEntropy(&TS);
  WriteCodebook(OutputFileName, &TS, Value(OverWrite));
  FreeCodebook(&TS);
  FreeBintree(&Tree);
  PrintTime(Watch);
  checkmemory();
}


